use std::ops::{Add, Mul, Neg};
use std::fmt;
use nalgebra::{Rotation3, Vector3};

#[derive(Clone, Serialize, Deserialize, Copy, Eq, PartialEq, Hash)]
pub struct V3 {
    pub x: i32,
    pub y: i32,
    pub z: i32,
}

impl V3 {
    pub fn new(x: i32, y: i32, z: i32) -> V3 {
        V3 { x, y, z }
    }

    pub fn neighbors(self) -> [V3; 6] {
        let o = self.offsets();
        [
            self + o[0],
            self + o[1],
            self + o[2],
            self + o[3],
            self + o[4],
            self + o[5],
        ]
    }

    pub fn rotate_x_90(self) -> V3 {
        self.rotate_90(Vector3::x_axis())
    }

    pub fn rotate_y_90(self) -> V3 {
        self.rotate_90(Vector3::y_axis())
    }

    pub fn rotate_z_90(self) -> V3 {
        self.rotate_90(Vector3::z_axis())
    }

    pub fn rotate_90(self, axis: nalgebra::Unit<nalgebra::Matrix<f64, nalgebra::base::dimension::U3, nalgebra::base::dimension::U1, nalgebra::ArrayStorage<f64, nalgebra::base::dimension::U3, nalgebra::base::dimension::U1>>>) -> V3 {
        let axis = Vector3::z_axis();
        let rot = Rotation3::from_axis_angle(&axis, std::f64::consts::FRAC_PI_2);
        let self_as_v = Vector3::new(self.x as f64, self.y as f64, self.z as f64);
        let tv = rot*self_as_v;
        V3::new(tv.x as i32, tv.y as i32, tv.z as i32)
    }

    pub fn rotated_neighbors(self, axis: u8,) -> [V3; 6] {

    //      [2]
    //   [1][S][0]
    //      [3]


        [
            self,
            self,
            self,
            self,
            self,
            self,
        ]
    }


    pub fn offsets(self) -> [V3; 6] {
        [
            V3::new(1, 0, 0),
            V3::new(-1, 0, 0),
            V3::new(0, 1, 0),
            V3::new(0, -1, 0),
            V3::new(0, 0, 1),
            V3::new(0, 0, -1),
        ]
    }
}

impl fmt::Debug for V3 {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        // // // write!(f, "V[{},{},{}]", self.x*8, self.y*8, self.z*8)
        write!(f, "V[{},{},{}]", self.x, self.y, self.z)
    }
}
impl Add<V3> for V3 {
    type Output = V3;
    fn add(self, other: V3) -> V3 {
        V3 {
            x: self.x + other.x,
            y: self.y + other.y,
            z: self.z + other.z,
        }
    }
}
impl Add<i32> for V3 {
    type Output = V3;
    fn add(self, other: i32) -> V3 {
        V3 {
            x: self.x + other,
            y: self.y + other,
            z: self.z + other,
        }
    }
}
impl Mul<i32> for V3 {
    type Output = V3;
    fn mul(self, other: i32) -> V3 {
        V3 {
            x: self.x * other,
            y: self.y * other,
            z: self.z * other,
        }
    }
}
impl Neg<> for V3 {
    type Output = V3;
    fn neg(self) -> V3 {
        V3 {
            x: -self.x,
            y: -self.y,
            z: -self.z,
        }
    }
}
impl<'a, 'b> Add<&'b V3> for &'a V3 {
    type Output = V3;

    fn add(self, other: &'b V3) -> V3 {
        V3 {
            x: self.x + other.x,
            y: self.y + other.y,
            z: self.z + other.z,
        }
    }
}
impl<'a> Mul<i32> for &'a V3 {
    type Output = V3;

    fn mul(self, other: i32) -> V3 {
        V3 {
            x: self.x * other,
            y: self.y * other,
            z: self.z * other,
        }
    }
}
impl<'a> Neg<> for &'a V3 {
    type Output = V3;

    fn neg(self) -> V3 {
        V3 {
            x: -self.x,
            y: -self.y,
            z: -self.z,
        }
    }
}

