#[macro_use]
extern crate serde_derive;
extern crate serde;
extern crate serde_json;
mod gridmath;
use std::fs::File;
use std::io::BufReader;
use std::io::BufWriter;
// use std::ops::{Add, Sub, Div};
use std::collections::HashMap;
#[macro_use] extern crate log;
extern crate simplelog;
use rand::seq::SliceRandom;
use simplelog::*;
use gridmath::*;


/// A graphics tile as ex- or imported from a 3d program
#[derive(Clone, Debug, Serialize, Deserialize)]
struct GfxTile {
    name: String,
    pos: Vec<f32>,
    #[serde(default)]
    rot: Vec<f32>
}

fn unitized_grid(vol: &HashMap<V3, String>, divisor: i32) -> HashMap<V3, String> {
    vol.iter()
        .map(|(k, v)| {
            (
                V3 {
                    x: k.x / divisor,
                    y: k.y / divisor,
                    z: k.z / divisor,
                },
                v.clone(),
            )
        })
        .collect()
}

fn guess_gridsize(vol: &HashMap<V3, String>) -> i32 {
    let mut x_positions: Vec<i32> = vol.iter().map(|(k, _v)| k.x).collect();
    let mut y_positions: Vec<i32> = vol.iter().map(|(k, _v)| k.y).collect();
    x_positions.sort();
    x_positions.dedup();
    y_positions.sort();
    y_positions.dedup();
    let mut dist_count: HashMap<i32, i32> = HashMap::new();
    let mut last: Option<i32> = None;
    for px in x_positions {
        if let Some(l) = last {
            let delta = px.abs() - l;
            let c = dist_count.entry(delta.abs()).or_insert(1);
            *c += 1;
        }
        last = Some(px.abs());
    }

    let mut grid = 1;
    let mut maxcount = 0;
    for (size, count) in dist_count {
        if count > maxcount {
            grid = size;
            maxcount = count;
        }
    }

    // dbg!(&grid);
    grid
}

fn train_volume(vol: &HashMap<V3, String>) -> HashMap<String, HashMap<V3, Vec<String>>> {
    let mut trained: HashMap<String, HashMap<V3, Vec<String>>> = HashMap::new();

    for (pos, tilename) in vol {
        // lookup neighbors

        let tile = trained.entry(tilename.clone()).or_insert(HashMap::new());

        for p_o in pos.offsets().iter() {
            let p_n = p_o + pos;
            if let Some(neighbor) = vol.get(&p_n) {
                if let Some(neighbor_tile) = tile.get_mut(&p_o) {
                    neighbor_tile.push(neighbor.clone());
                } else {
                    tile.insert(p_o.clone(), vec![neighbor.clone()]);
                }
            }
        }
    }
    trained
}

fn load_export(path: &'static str) -> HashMap<V3, String> {
    let reader = BufReader::new(File::open(path).unwrap());
    let loaded_tiles: Vec<GfxTile> = serde_json::from_reader(reader).expect("Could not parse JSON");
    loaded_tiles
        .iter()
        .map(|t| {
            dbg!(&t);
            (
                V3::new(t.pos[0] as i32, t.pos[1] as i32, t.pos[2] as i32),
                t.name.clone(),
            )
        })
        .collect()
}



fn get_unsolved_neighbors(cell: &V3, grid: &HashMap<V3, String>) -> Vec<V3> {
    let mut neighbors = vec![];

    for n in &cell.neighbors() {
        if !grid.contains_key(n) {
            neighbors.push(n.clone());
        }
    }
    neighbors
}

fn solve(constraint: &HashMap<V3, String>, training_data: &HashMap<String, HashMap<V3, Vec<String>>>) -> Option<String> {
    
 
    let mut valid_templates = vec![];
    for template in training_data {
        
        let c: Vec<bool> = constraint
        .iter()
        .map(|(pos, name)| template.1.get(&pos).map(|x| x.contains(name)) )
        .filter(|e| e.is_some())
        .map(|e| e.unwrap())
        .collect();

        if c.is_empty() {
            valid_templates.push(None);
        } else {
            if c.iter().all(|x| x == &true) {
                valid_templates.push(Some(template.0))
            } 
        }
    
        // valid_templates.push(c);

    }


    let valid_tiles: Vec<_> = valid_templates.iter().filter(|x| x.is_some()).map(|x| x.unwrap()).collect();

    // This was not solvable
    if valid_tiles.is_empty() {
        return None
    } else {

        Some(valid_tiles.choose(&mut rand::thread_rng()).unwrap().to_string())
        // Some("grass".to_string())
    }
    


}

fn wfc_solver(
    training_data: &HashMap<String, HashMap<V3, Vec<String>>>,
    base_map: HashMap<V3, String>,
) -> HashMap<V3, String> {
    let mut new_grid = base_map;


    for (pos, _name) in new_grid.clone() {
        for unsolved_neighbor in get_unsolved_neighbors(&pos, &new_grid) {

            debug!("Unsolved {:?}", unsolved_neighbor);
            
            let mut contraint_map: HashMap<V3, String> = HashMap::new();
            
            for tile_neighbor in &unsolved_neighbor.offsets() {
                let abs_neighbor_pos = &unsolved_neighbor + tile_neighbor;

                if let Some(tn) = new_grid.get(&abs_neighbor_pos) {
                    debug!("\tNeighbor {:?} exists, is {}. Inserting as constraint.", abs_neighbor_pos, tn);
                    contraint_map.insert(tile_neighbor.clone(), tn.to_string());
                }
            }
            // Constraint map is ready here      
            debug!("\tSolving {} contraints in {:?}", contraint_map.len(), contraint_map );


            match solve(&contraint_map, &training_data) {
                Some(tile) => {
                    new_grid.insert(unsolved_neighbor, tile);
                },
                None => info!("Unsolved tile")
            };
           
        }
    }

    // println!("grid: {:?}", new_grid);

    new_grid
}

fn main() {


    let myvec = V3::new(1, 0, 0);
    dbg!(&myvec);
    dbg!(myvec.rotate_z_90());


    // return;


    let _ = simplelog::SimpleLogger::init(LevelFilter::Info, Config::default());

    let mut import_volume: HashMap<V3, String> = HashMap::new();
    let imported_tiles = load_export("out.json");
    let gridsize = guess_gridsize(&imported_tiles);
    info!("Detected grid size: {}", gridsize);
    let imported_tiles = unitized_grid(&imported_tiles, gridsize);
    info!("Tiles imported: {}", imported_tiles.len());

    for (pos, tile) in imported_tiles {
        import_volume.insert(pos, tile);
    }

    let training_data = train_volume(&import_volume);
    info!("Training done: {} tiles", training_data.len());

    let mut base_map: HashMap<V3, String> = HashMap::new();

    // init start conditions
    base_map.insert(V3::new(0, 0, 0), String::from("grass"));

    for i in 1..8 {
        info!(
            "=========================== Iteration {} ===========================",
            i
        );
        base_map = wfc_solver(&training_data, base_map);
    }


    let mut export = vec![];
    for (p, n) in &base_map {
        export.push((p * gridsize, n));
    }
    let writer = BufWriter::new(File::create("map.json").unwrap());
    serde_json::to_writer_pretty(writer, &export).unwrap();
}
